from __future__ import unicode_literals

import json
from random import randint
from mock import patch

from django.test import SimpleTestCase, override_settings
from django.core.urlresolvers import reverse_lazy

from .utils import AnonymizerServiceTestUtils
from core.models import UniqueVote, ElectionVote
from . import AnonymizerServiceTestCase


class AnonymizeVoteTest(AnonymizerServiceTestUtils,
                        AnonymizerServiceTestCase):
    """
    Class to test the anonymizeVote API endpoint
    """
    # Setup the Test Case
    def setUp(self):
        """Setup all needed endpoints."""
        self.election_id = 34
        req_kwargs = {'election_id': self.election_id}
        self.url = reverse_lazy(
            'api:anonymizevote',
            kwargs=req_kwargs
        )
        self.init_url = reverse_lazy(
            'api:initializeelection',
            kwargs=req_kwargs
        )

    def test_empty_request(self):
        """Test for empty request."""
        r = self.client.post(self.url)
        self.assertEqual(r.status_code, 400)

    def test_not_existing_election(self):
        """Submit a vote in a non-existing election."""
        vote_obj = self.create_vote_obj(self.election_id)
        r = self.client.post(self.url, vote_obj)
        self.assertEqual(r.status_code, 400)

    def test_full_slots(self):
        """
        Force the maximum number of votes to be reached
        """
        # Initialize election
        election_id = self.election_id
        init_obj = self.create_init(2)
        self.client.post(self.init_url, init_obj)

        # Submit valid vote
        vote_obj = self.create_vote_obj(election_id)
        r = self.client.post(self.url, vote_obj)
        self.assertEqual(r.status_code, 200)

        # Submit another valid vote
        vote_obj = self.create_vote_obj(election_id)
        r = self.client.post(self.url, vote_obj)
        self.assertEqual(r.status_code, 200)

        # Maximum number of votes reached exception
        # must be raised
        vote_obj = self.create_vote_obj(election_id)
        r = self.client.post(self.url, vote_obj)
        response = r.content.decode('string-escape').strip('"')
        self.assertEqual(response, "false")

    def test_vote_submit(self):
        """Test a full vote submission workflow."""
        election_id = self.election_id
        init_obj = self.create_init(3)
        self.client.post(self.init_url, init_obj)
        vote_obj = self.create_vote_obj(election_id)
        r = self.client.post(self.url, vote_obj)
        self.assertEqual(r.status_code, 200)

    @override_settings(EVOTE_OPTIMIZED_INSERT=True)
    @patch('core.utils.get_redis_connection')
    def test_optimized_vote_submit(self, mock_redis_conn):
        """Test a full vote submission workflow."""
        mock_redis_conn.return_value.smembers.return_value = ["1,10"]
        mock_redis_conn.return_value.sadd.side_effect = [False, True, True]
        # self.sadd_side_effect
        election_id = self.election_id
        init_obj = self.create_init(3)
        self.client.post(self.init_url, init_obj)
        vote_obj = self.create_vote_obj(election_id)
        r = self.client.post(self.url, vote_obj)
        self.assertEqual(r.status_code, 200)


class OperateElectionTest(AnonymizerServiceTestUtils,
                          AnonymizerServiceTestCase):
    """
    Class to test all operations with election table API endpoints
    """
    # Setup the Test Case
    def setUp(self):
        """Setup all needed endpoints."""
        self.dummy_election_id = str(randint(1, 10))
        req_kwargs = {'election_id': self.dummy_election_id}
        self.init_url = reverse_lazy(
            'api:initializeelection',
            kwargs=req_kwargs
        )
        self.close_url = reverse_lazy(
            'api:closeelection',
            kwargs=req_kwargs
        )
        self.election_root_url = reverse_lazy('api:electionendpoint')
        invalid_req_kwargs = {'election_id': 'string'}
        self.invalid_init_url = reverse_lazy(
            'api:initializeelection',
            kwargs=invalid_req_kwargs
        )

    def test_election_no_pk(self):
        """Test for empty election_id on request."""
        r = self.client.post(self.election_root_url)
        self.assertEqual(r.status_code, 405)

    def test_empty_request_init(self):
        """Test for empty request on election init"""
        r = self.client.post(self.init_url)
        self.assertEqual(r.status_code, 200)

    def test_empty_request_close(self):
        """Test for empty request on election close."""
        r = self.client.post(self.close_url)
        self.assertEqual(r.status_code, 400)

    def test_not_integer_voters(self):
        """Test when number of voters is not a proper integer."""
        init_obj = self.create_init("string")
        r = self.client.post(self.init_url, init_obj)
        self.assertEqual(r.status_code, 200)

    def test_election_initialization(self):
        """Test a successful election initialization."""
        init_obj = self.create_init(20)
        r = self.client.post(self.init_url, init_obj)
        self.assertEqual(r.status_code, 200)

    @override_settings(EVOTE_BATCH_LIMIT=5)
    def test_batched_election_initialization(self):
        """Test a successful election initialization."""
        init_obj = self.create_init(20)
        r = self.client.post(self.init_url, init_obj)
        self.assertEqual(r.status_code, 200)
        json_response = json.loads(r.content)
        self.assertEqual(json_response["error"], False)

    def test_duplicate_election_initialization(self):
        """Test a successful election initialization."""
        init_obj = self.create_init(20)
        r = self.client.post(self.init_url, init_obj)
        self.assertEqual(r.status_code, 200)
        r = self.client.post(self.init_url, init_obj)
        json_response = json.loads(r.content)
        self.assertEqual(json_response["error"], True)

    @patch('core.views_api.initialize_election')
    def test_failed_election_initialization(self, mock_fn):
        """Test a successful election initialization."""
        mock_fn.return_value = False
        init_obj = self.create_init(20)
        r = self.client.post(self.init_url, init_obj)
        json_response = json.loads(r.content)
        self.assertEqual(r.status_code, 200)
        self.assertEqual(json_response["error"], True)

    def test_election_close(self):
        """Test a successful election close."""
        init_obj = self.create_init(20)
        r = self.client.post(self.init_url, init_obj)
        self.assertEqual(r.status_code, 200)
        r = self.client.post(self.close_url)
        self.assertEqual(r.status_code, 200)
        json_response = json.loads(r.content)
        self.assertTrue(json_response)

    def test_election_table_entries(self):
        """
        Test if election is initialized with the requested number of entries.
        """
        voters_nr = 10
        init_obj = self.create_init(voters_nr)
        r = self.client.post(self.init_url, init_obj)
        self.assertEqual(r.status_code, 200)
        self.assertTrue(
            self.check_table_integrity(self.dummy_election_id, voters_nr)
        )

    def test_election_integrity_entries(self):
        """
        Test if election is initialized with the requested number of entries.
        """
        voters_nr = 1
        init_obj = self.create_init(voters_nr)
        r = self.client.post(self.init_url, init_obj)
        self.assertEqual(r.status_code, 200)
        for i in range(1, 3):
            ElectionVote(election_id=self.dummy_election_id).save()
        self.assertFalse(
            self.check_table_integrity(self.dummy_election_id, voters_nr)
        )
        for i in range(1, 3):
            UniqueVote(election_id=self.dummy_election_id).save()
        self.assertFalse(
            self.check_table_integrity(self.dummy_election_id, voters_nr)
        )


class ListReferenceElectionTest(AnonymizerServiceTestUtils,
                                AnonymizerServiceTestCase):
    """
    Class to test all to the list reference API endpoint
    """
    # Setup the Test Case
    def setUp(self):
        """Setup all needed endpoints."""
        # Random election ID
        self.dummy_election_id = str(randint(1, 10))
        # Random unique vote reference
        self.dummy_reference = "Y#7kslpsfa3_2iu"
        req_kwargs = {'election_id': self.dummy_election_id}
        # Initialize election URL
        self.init_url = reverse_lazy(
            'api:initializeelection',
            kwargs=req_kwargs
        )
        # Reference list URL
        self.references_url = reverse_lazy(
            'api:listreferences',
            kwargs=req_kwargs
        )
        # Insert vote URL
        self.vote_url = reverse_lazy(
            'api:anonymizevote',
            kwargs=req_kwargs
        )

    def test_references_post_method(self):
        """Test for method POST not allowed."""
        r = self.client.post(self.references_url)
        self.assertEqual(r.status_code, 405)

    def test_references_get_method(self):
        """Test empty list of references for non-existing election."""
        r = self.client.get(self.references_url)
        self.assertEqual(r.status_code, 200)
        response = r.content.decode('string-escape').strip('"')
        json_response = json.loads(response)
        # Check that there are no listed references
        # for elections not initialized
        self.assertEqual(json_response["count"], 0)

    def test_references_inserted_vote(self):
        """
        Test pagination and inserted vote count for a single bulletin.
        """
        election_id = self.dummy_election_id
        # Initialize election
        init_obj = self.create_init(3)
        r = self.client.post(self.init_url, init_obj)
        self.assertEqual(r.status_code, 200)

        # Insert Vote
        vote_obj = self.create_vote_obj(election_id)
        r = self.client.post(self.vote_url, vote_obj)
        self.assertEqual(r.status_code, 200)

        # Get references list
        r = self.client.get(self.references_url)
        self.assertEqual(r.status_code, 200)
        response = r.content.decode('string-escape').strip('"')
        json_response = json.loads(response)

        # Check if reference was fetched
        self.assertEqual(json_response["count"], 1)

        # Check that there is no next results page
        self.assertIsNone(json_response["next"])

    def test_references_paginated_vote(self):
        """
        Test pagination and inserted vote count for multiple bulletins.
        """
        election_id = self.dummy_election_id
        # Initialize election for 25 voters
        init_obj = self.create_init(25)
        r = self.client.post(self.init_url, init_obj)

        # Insert 21 votes
        for i in range(1, 22):
            vote_obj = self.create_vote_obj(election_id)
            r = self.client.post(self.vote_url, vote_obj)

        # List all the existing references
        r = self.client.get(self.references_url)
        self.assertEqual(r.status_code, 200)
        response = r.content.decode('string-escape').strip('"')
        json_response = json.loads(response)

        # Check if the number of votes matches
        self.assertEqual(json_response["count"], 21)

        # Since the pagination is 20
        # a next page URL must exist
        self.assertIsNotNone(json_response["next"])

    def test_references_multiple_votes(self):
        """
        Test pagination and inserted vote count for multiple bulletins.
        """
        election_id = self.dummy_election_id
        # Initialize election for 25 voters
        init_obj = self.create_init(25)
        r = self.client.post(self.init_url, init_obj)

        # Insert 21 votes
        for i in range(1, 22):
            vote_obj = self.create_vote_obj(election_id)
            r = self.client.post(self.vote_url, vote_obj)

        # Fetch the 2nd page of references
        r = self.client.get(self.references_url, {'page': 2})

        # Check if the page exists
        self.assertEqual(r.status_code, 200)
        response = r.content.decode('string-escape').strip('"')
        json_response = json.loads(response)

        # Check if the number of votes matches
        self.assertEqual(len(json_response["results"]), 1)

    def test_filter_vote_reference(self):
        """Test filtering for a non-existing vote reference."""
        election_id = self.dummy_election_id
        vote_reference = self.dummy_reference
        # Initialize election
        init_obj = self.create_init(2)
        self.client.post(self.init_url, init_obj)

        # Insert one random vote reference
        vote_obj = self.create_vote_obj(election_id)
        self.client.post(self.vote_url, vote_obj)

        # Filter for a non-existing reference
        r = self.client.get(self.references_url,
                            {'unique_vote_bulletin': vote_reference})

        self.assertEqual(r.status_code, 200)
        response = r.content.decode('string-escape').strip('"')
        json_response = json.loads(response)

        # Check that no such reference exists
        self.assertEqual(json_response["count"], 0)

    def test_filter_existing_vote_reference(self):
        """Test filtering for an existing vote reference."""
        election_id = self.dummy_election_id
        vote_reference = self.dummy_reference
        # Initialize election
        init_obj = self.create_init(2)
        self.client.post(self.init_url, init_obj)
        # Insert one vote with the dummy reference
        vote_obj = self.create_vote_obj(election_id)
        vote_obj['unique_bulletin'] = vote_reference
        self.client.post(self.vote_url, vote_obj)

        # Filter references that match the provided dummy
        r = self.client.get(self.references_url,
                            {'unique_vote_bulletin': vote_reference})

        self.assertEqual(r.status_code, 200)
        response = r.content.decode('string-escape').strip('"')
        json_response = json.loads(response)
        # Check that a reference exists
        self.assertEqual(json_response["count"], 1)

        reference_list = json_response["results"][0]
        # And it is the provided reference
        self.assertEqual(reference_list['unique_vote_bulletin'],
                         vote_reference)


class ListElectionVotesTest(AnonymizerServiceTestUtils,
                            AnonymizerServiceTestCase):
    """
    Class to test all to the list reference API endpoint
    """
    # Setup the Test Case
    def setUp(self):
        """Setup all needed endpoints."""
        # Random election ID
        self.dummy_election_id = str(randint(1, 10))
        req_kwargs = {'election_id': self.dummy_election_id}
        # Initialize election URL
        self.init_url = reverse_lazy(
            'api:initializeelection',
            kwargs=req_kwargs
        )
        # Reference list URL
        self.votes_url = reverse_lazy(
            'api:listvotes',
            kwargs=req_kwargs
        )
        # Insert vote URL
        self.vote_url = reverse_lazy(
            'api:anonymizevote',
            kwargs=req_kwargs
        )

    def test_votes_post_method(self):
        """Test for method POST not allowed."""
        r = self.client.post(self.votes_url)
        self.assertEqual(r.status_code, 405)

    def test_votes_get_method(self):
        """Test empty list of references for non-existing election."""
        r = self.client.get(self.votes_url)
        self.assertEqual(r.status_code, 200)
        response = r.content.decode('string-escape').strip('"')
        json_response = json.loads(response)
        # Check that there are no listed votes
        # for elections not initialized
        self.assertEqual(json_response["count"], 0)

    def test_votes_inserted_vote(self):
        """Test pagination and inserted vote count."""
        election_id = self.dummy_election_id
        # Initialize election
        init_obj = self.create_init(3)
        r = self.client.post(self.init_url, init_obj)
        self.assertEqual(r.status_code, 200)

        # Insert Vote
        vote_obj = self.create_vote_obj(election_id)
        r = self.client.post(self.vote_url, vote_obj)
        self.assertEqual(r.status_code, 200)

        # Get vote list
        r = self.client.get(self.votes_url)
        self.assertEqual(r.status_code, 200)
        response = r.content.decode('string-escape').strip('"')
        json_response = json.loads(response)

        # Check if vote was fetched
        self.assertEqual(json_response["count"], 1)

        # Check that there is no next results page
        self.assertIsNone(json_response["next"])

    def test_votes_paginated_vote(self):
        """Test pagination and inserted multiple votes."""
        election_id = self.dummy_election_id
        # Initialize election for 25 voters
        init_obj = self.create_init(25)
        r = self.client.post(self.init_url, init_obj)

        # Insert 21 votes
        for i in range(1, 22):
            vote_obj = self.create_vote_obj(election_id)
            r = self.client.post(self.vote_url, vote_obj)

        # List all the existing votes
        r = self.client.get(self.votes_url)
        self.assertEqual(r.status_code, 200)
        response = r.content.decode('string-escape').strip('"')
        json_response = json.loads(response)

        # Check if the number of votes match
        self.assertEqual(json_response["count"], 21)

        # Since the pagination is 20
        # a next page URL must exist
        self.assertIsNotNone(json_response["next"])

    def test_votes_multiple(self):
        """Test pagination and inserted vote count."""
        election_id = self.dummy_election_id
        # Initialize election for 25 voters
        init_obj = self.create_init(25)
        r = self.client.post(self.init_url, init_obj)

        # Insert 21 votes
        for i in range(1, 22):
            vote_obj = self.create_vote_obj(election_id)
            r = self.client.post(self.vote_url, vote_obj)

        # Fetch the 2nd page of votes
        r = self.client.get(self.votes_url, {'page': 2})

        # Check if the page exists
        self.assertEqual(r.status_code, 200)
        response = r.content.decode('string-escape').strip('"')
        json_response = json.loads(response)

        # Check if the number of votes matches
        self.assertEqual(len(json_response["results"]), 1)


class TestPingView(SimpleTestCase):
    """"""
    def setUp(self):
        """Setup all needed endpoints."""
        self.url = reverse_lazy('api:ping')

    def test_server_alive(self):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
